@@ -34,7 +34,9 @@
 
 #include "emil.h"
 #include "charset/charset.h"
+int     strptrncnv(CHARSET *, CHARSET *, struct data *, struct data *, int);
 
+static char * out_data_part(struct data *);
 
 static char delim[256] = {
 0, 0, 0, 0, 0, 0, 0, 0,  0, 1, 1, 0, 0, 1, 0, 0,  
@@ -121,7 +123,7 @@ int parse_rfc1522(struct header *th)
   struct hprs *h;
   char *field;
   int i;
-  char *c;
+
   if (th->value == NULL)
     return(OK);
 #ifdef DEBUG
@@ -176,12 +178,7 @@ parse_hline(struct hprs *h)
   int type = UNKNOWN;
   char *inb;
   int len;
-  char *tmp;
 
-#ifdef DEBUG
-  if (edebug)
-    fprintf(stderr, "+   Parse hline.\n");
-#endif
   inbuf = h->sd;
   inb = inbuf->contents + inbuf->offset;
   offset = inbuf->offset;
@@ -225,7 +222,7 @@ parse_hline(struct hprs *h)
 	  inbuf->bodyend = inbuf->bodystart + len;
 #ifdef DEBUG
 	  if (edebug)
-	    fprintf(stderr, "-   Found comment.\n");
+	    fprintf(stderr, "-   Found comment '%s'.\n", out_data_part(inbuf));
 #endif
 	  tmpd = copy_hprs(h);
 	  tmpd->sd->end = inbuf->bodyend ? inbuf->bodyend : inbuf->end;
@@ -279,7 +276,7 @@ parse_hline(struct hprs *h)
 	  inbuf->bodyend = inbuf->bodystart + len;
 #ifdef DEBUG
 	  if (edebug)
-	    fprintf(stderr, "-   Found qstring.\n");
+	    fprintf(stderr, "-   Found qstring '%s'.\n", out_data_part(inbuf));
 #endif
 	  tmpd = copy_hprs(h);
 	  tmpd->types = ATOM;
@@ -334,7 +331,7 @@ parse_hline(struct hprs *h)
 	  inbuf->bodyend = inbuf->bodystart + len;
 #ifdef DEBUG
 	  if (edebug)
-	    fprintf(stderr, "-   Found domain literal.\n");
+	    fprintf(stderr, "-   Found domain literal '%s'.\n", out_data_part(inbuf));
 #endif
 	  tmpd = copy_hprs(h);
 	  tmpd->types = ATOM;
@@ -389,7 +386,7 @@ parse_hline(struct hprs *h)
 	  inbuf->bodyend = inbuf->bodystart + len;
 #ifdef DEBUG
 	  if (edebug)
-	    fprintf(stderr, "-   Found address.\n");
+	    fprintf(stderr, "-   Found address '%s'.\n", out_data_part(inbuf));
 #endif
 	  tmpd = copy_hprs(h);
 	  tmpd->types = ATOM;
@@ -464,6 +461,10 @@ parse_hline(struct hprs *h)
 	  inb++;
 	  if (*inb != '?')
 	    {
+#ifdef DEBUG
+	      if (edebug)
+		fprintf(stderr, "*   rfc1522 decoding failed on starting '=?'\n");
+#endif
 	      type = type <<1;
 	      inbuf->offset = offset;
 	      break;
@@ -473,6 +474,10 @@ parse_hline(struct hprs *h)
 	  /* Get charset */
 	  if ((len = where(inb, '?')) == 0)
 	    {
+#ifdef DEBUG
+	      if (edebug)
+		fprintf(stderr, "*   rfc1522 decoding failed on second '='\n");
+#endif
 	      type = type <<1;
 	      inbuf->offset = offset;
 	      break;
@@ -498,6 +503,10 @@ parse_hline(struct hprs *h)
 	    }
 	  if (inbuf->encoding == 0)
 	    {
+#ifdef DEBUG
+	      if (edebug)
+		fprintf(stderr, "*   rfc1522 decoding failed on encoding '%c'\n", *inb);
+#endif
 	      type = type <<1;
 	      inbuf->offset = offset;
 	      break;
@@ -506,6 +515,10 @@ parse_hline(struct hprs *h)
 	  inbuf->offset += 1;
 	  if (*inb != '?')
 	    {
+#ifdef DEBUG
+	      if (edebug)
+		fprintf(stderr, "*   rfc1522 decoding failed on third '?'\n");
+#endif
 	      type = type <<1;
 	      inbuf->offset = offset;
 	      break;
@@ -516,12 +529,20 @@ parse_hline(struct hprs *h)
 	  /* Get string */
 	  if ((len = where(inb, '?')) == 0)
 	    {
+#ifdef DEBUG
+	      if (edebug)
+		fprintf(stderr, "*   rfc1522 decoding failed on string\n");
+#endif
 	      type = type <<1;
 	      inbuf->offset = offset;
 	      break;
 	    }
 	  if (*(inbuf->contents + len + inbuf->offset) != '?')
 	    {
+#ifdef DEBUG
+	      if (edebug)
+		fprintf(stderr, "*   rfc1522 decoding failed on closing '?'\n");
+#endif
 	      type = type <<1;
 	      inbuf->offset = offset;
 	      break;
@@ -529,6 +550,10 @@ parse_hline(struct hprs *h)
 	  inb += len + 1;
 	  if (*inb != '=')
 	    {
+#ifdef DEBUG
+	      if (edebug)
+		fprintf(stderr, "*   rfc1522 decoding failed on closing '='\n");
+#endif
 	      type = type <<1;
 	      inbuf->offset = offset;
 	      break;
@@ -541,7 +566,7 @@ parse_hline(struct hprs *h)
 	  inb++;
 #ifdef DEBUG
 	  if (edebug)
-	    fprintf(stderr, "-   Found rfc1522 encoded string.\n");
+	    fprintf(stderr, "-   Found rfc1522 encoded string '%s'.\n", out_data_part(inbuf));
 #endif
 	  tmpd = copy_hprs(h);
 	  tmpd->types = h->types;
@@ -563,7 +588,7 @@ parse_hline(struct hprs *h)
 	case ATOM:
 	  offset = inbuf->offset;
 	  h->types = h->types | DELIMITER;
-	  if (delim[*inb])
+	  if (delim[*inb] != '\0')
 	    {
 	      type = type <<1;
 	      inbuf->offset = offset;
@@ -574,7 +599,7 @@ parse_hline(struct hprs *h)
 	      inbuf->bodystart = inbuf->offset;
 	      while(inbuf->offset < inbuf->end)
 		{
-		  if (delim[*inb])
+		  if (delim[*inb] != '\0')
 		    break;
 		  else
 		    {
@@ -585,7 +610,7 @@ parse_hline(struct hprs *h)
 	      inbuf->bodyend = inbuf->offset;
 #ifdef DEBUG
 	  if (edebug)
-	    fprintf(stderr, "-   Found atom.\n");
+	    fprintf(stderr, "-   Found atom '%s'.\n", out_data_part(inbuf));
 #endif
 	      tmpd = copy_hprs(h);
 	      tmpd->types = h->types;
@@ -607,27 +632,20 @@ parse_hline(struct hprs *h)
 	  
 	case DELIMITER:
 	  offset = inbuf->offset;
-	  if (!delim[*inb])
+	  if (delim[*inb] == '\0')
 	    {
 	      type = type <<1;
 	      inbuf->offset = offset;
 	      break;
 	    }
-	  else
-	    {
-	      inbuf->bodystart = inbuf->offset;
-	      while(inbuf->offset < inbuf->end)
-		{
-		  if (delim[*inb])
-		    {
-		      inb++;
-		      inbuf->offset += 1;
-		    }
-		  else
-		    break;
-		}
-	      inbuf->bodyend = inbuf->offset;
-	    }
+	  inbuf->bodystart = inbuf->offset;
+	  inb++;
+	  inbuf->offset += 1;
+	  inbuf->bodyend = inbuf->offset;
+#ifdef DEBUG
+	  if (edebug)
+	    fprintf(stderr, "-   Found delimiter '%s'.\n", out_data_part(inbuf));
+#endif
 	  tmpd = copy_hprs(h);
 	  tmpd->types = h->types;
 	  
@@ -680,6 +698,7 @@ convert_hlines(struct hprs *h)
   return(OK);
 }
 
+int
 recode_hline(struct hprs *h)
 {
   switch (h->td->encoding)
@@ -690,7 +709,8 @@ recode_hline(struct hprs *h)
     fprintf(stderr, "-   is 7bit\n");
 #endif
       if (h->td->charset == NULL)
-	h->td->charset = NEWSTR("US-ASCII");
+	h->td->charset = source->charset == NULL ? NEWSTR("ISO-8859-1"):
+	  source->charset;
       if (h_tocharset(h) == OK)
 	check_bits(h->td);
 
@@ -782,7 +802,6 @@ h_tocharset(struct hprs *h)
   CHARSET *schar, *tchar;
   struct data *inbuf, *outbuf;
   char *sset, *tset;
-  long offset;
   int copy = 0;
 
   inbuf = h->td;
@@ -797,6 +816,14 @@ h_tocharset(struct hprs *h)
   if (edebug)
     fprintf(stderr, "* Header tocharset");
 #endif
+  if (inbuf->size == 0)
+    {
+#ifdef DEBUG
+      if (edebug)
+	fprintf(stderr, "--- Empty input (failed).\n");
+#endif
+      return(NOK);
+    }
   if (sset == NULL || tset == NULL ||
       (schar = (CHARSET *)getchset(sset, 29)) == NULL ||
       (tchar = (CHARSET *)getchset(tset, 29)) == NULL)
@@ -818,7 +845,7 @@ h_tocharset(struct hprs *h)
     fprintf(stderr, ", from %s to %s", sset, tset);
 #endif
       outbuf = (struct data *) Yalloc(sizeof(struct data ));
-      strptrncnv(tchar, schar, outbuf, inbuf);
+      strptrncnv(tchar, schar, outbuf, inbuf, HEAD_BUF);
       outbuf->bodyend = outbuf->end;
       outbuf->charset = NEWSTR(tset);
       h->td = outbuf;
@@ -838,3 +865,12 @@ h_tocharset(struct hprs *h)
 #endif
   return(OK);
 }
+
+char *
+out_data_part(struct data *d)
+{
+  char *c;
+  c = (char *)Yalloc(2 + d->bodyend - d->bodystart);
+  strncpy(c, d->contents + d->bodystart, d->bodyend - d->bodystart);
+  return(c);
+}
